package org.xbib.multiserv.plugin.tanserv;

import java.io.IOException;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.time.Instant;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.xbib.multiserv.api.Command;
import org.xbib.multiserv.api.CommandFunction;
import org.xbib.multiserv.api.CommandServer;
import org.xbib.multiserv.plugin.jdbc.JDBCPlugin;

public class TanServ extends JDBCPlugin {

    private static final Logger logger = Logger.getLogger(TanServ.class.getName());

    public TanServ() {
        super();
    }

    @Override
    public void onStart() throws IOException {
        super.onStart();
        try {
            executeStatement("create table if not exists tan (\"tan\" varchar(8), \"owner\" varchar(22), \"grp\" varchar(22), \"value\" varchar(2), \"created\" timestamp, \"deleted\" timestamp)", Map.of());
        } catch (SQLException e) {
            throw new IOException("unable to create table tan");
        }
        logger.log(Level.INFO, "started");
    }

    @Override
    public List<CommandFunction> getFunctions(CommandServer commandServer) {
        List<CommandFunction> list = new ArrayList<>();
        list.add(CommandFunction.of("SLNPCreateTAN", this, SLNPCreateTAN.class));
        list.add(CommandFunction.of("SLNPInsertTAN", this, SLNPInsertTAN.class));
        list.add(CommandFunction.of("SLNPDeleteTAN", this, SLNPDeleteTAN.class));
        list.add(CommandFunction.of("SLNPShowTAN", this, SLNPShowTAN.class));
        list.add(CommandFunction.of("SLNPIsTanValid", this, SLNPIsTanValid.class));
        list.add(CommandFunction.of("SLNPShowTanGroup", this, SLNPShowTanGroup.class));
        return list;
    }

    @Override
    public void onExecute(CommandServer commandServer, Command<?> command, Map<String, Object> params, List<String> response) {
    }

    @Override
    public void onShutdown() {
        super.onShutdown();
    }

    private List<String> createTAN(String owner, String value, String group, int count) {
        if (count < 1) {
            throw new IllegalArgumentException("too few TANs requested");
        }
        if (count > 1000) {
            throw new IllegalArgumentException("too many TANs requested");
        }
        int tries = 4 * count;
        Set<String> set = new LinkedHashSet<>();
        while (set.size() < count && tries-- > 0) {
            String tan = RandomTAN.createPseudoTAN(7, true, true);
            if (set.contains(tan)) {
                continue;
            }
            if (insertTAN(owner, value, group, tan)) {
                set.add(tan);
            }
        }
        logger.log(Level.INFO, count + " new TANs created (" + (4 * count - tries) + " tries) for " + owner + ": " + set);
        return new ArrayList<>(set);
    }

    public Map<String, Object> isTANValid(String owner, String value, String group, String tan) throws SQLException {
        if (tan != null && owner != null && group != null && value != null) {
            return getRow("select \"created\", \"value\", \"grp\" from tan where \"tan\" = :tan and \"owner\" = :owner and \"grp\" = :grp and \"value\" = :value and \"deleted\" IS NULL",
                    Map.of("tan", tan, "owner", owner, "grp", group, "value", value));
        } else if (tan != null && owner != null && group != null) {
            return getRow("select \"created\", \"value\", \"grp\" from tan where \"tan\" = :tan and \"owner\" = :owner and \"grp\" = :grp and \"deleted\" IS NULL",
                    Map.of("tan", tan, "owner", owner, "grp", group));
        } else if (tan != null && owner != null) {
            return getRow("select \"created\", \"value\", \"grp\" from tan where \"tan\" = :tan and \"owner\" = :owner and \"deleted\" IS NULL",
                    Map.of("tan", tan, "owner", owner));
        } else if (tan != null) {
            return getRow("select \"created\", \"value\", \"grp\" from tan where \"tan\" = :tan and \"deleted\" IS NULL",
                    Map.of("tan", tan));
        }
        return null;
    }

    private Map<String, Object> showTAN(String owner, String tan) throws SQLException {
        return owner != null ?
                getRow("select * from tan where \"tan\" = :tan and \"owner\" = :owner",
                        Map.of("tan", tan, "owner", owner)) :
                getRow("select * from tan where \"tan\" = :tan",
                        Map.of("tan", tan));
    }

    private boolean insertTAN(String owner, String value, String group, String tan) {
        if (owner == null) {
            throw new IllegalArgumentException("TAN-Sigel not specified");
        }
        if (tan == null) {
            throw new IllegalArgumentException("TAN not specified");
        }
        try {
            Timestamp timestamp = new Timestamp(Instant.now().toEpochMilli());
            if (group == null && value == null) {
                insertStatement("insert into tan(\"tan\", \"owner\", \"created\") values(:tan, :owner, :created)",
                        Map.of("tan", tan, "owner", owner, "created", timestamp));
            } else if (group == null) {
                insertStatement("insert into tan(\"tan\", \"owner\", \"value\", \"created\") values(:tan, :owner, :value, :created)",
                        Map.of("tan", tan, "owner", owner, "value", value, "created", timestamp));
            } else {
                insertStatement("insert into tan(\"tan\", \"owner\", \"grp\", \"value\", \"created\") values(:tan, :owner, :grp, :value, :created)",
                        Map.of("tan", tan, "owner", owner, "grp", group, "value", value, "created", timestamp));
            }
        } catch (Exception e) {
            // primary key exceptions etc.
            logger.log(Level.SEVERE, e.getMessage(), e);
            return false;
        }
        return true;
    }

    private List<String> getAllValidTANs(String owner, String value, String group) {
        List<String> rows = new ArrayList<>();
        try {
            if (value != null && group != null) {
                queryAndApply("select \"tan\" from tan where \"owner\" = :owner and \"grp\" = :grp and \"value\" = :value",
                        Map.of("owner", owner, "grp", group, "value", value),
                        r -> {
                            try {
                                while (r.next()) {
                                    rows.add(r.getString("tan"));
                                }
                            } catch (SQLException t) {
                                logger.log(Level.SEVERE, t.getMessage(), t);
                            }
                        });
            } else if (group != null) {
                queryAndApply("select \"tan\" from tan where \"owner\" = :owner and \"grp\" = :grp",
                        Map.of("owner", owner, "grp", group),
                        r -> {
                            try {
                                while (r.next()) {
                                    rows.add(r.getString("tan"));
                                }
                            } catch (SQLException t) {
                                logger.log(Level.SEVERE, t.getMessage(), t);
                            }
                        });
            } else if (value != null) {
                queryAndApply("select \"tan\" from tan where \"owner\" = :owner and \"value\" = :value",
                        Map.of("owner", owner, "value", value),
                        r -> {
                            try {
                                while (r.next()) {
                                    rows.add(r.getString("tan"));
                                }
                            } catch (SQLException t) {
                                logger.log(Level.SEVERE, t.getMessage(), t);
                            }
                        });
            } else {
                queryAndApply("select \"tan\" from tan where \"owner\" = :owner",
                        Map.of("owner", owner),
                        r -> {
                            try {
                                while (r.next()) {
                                    rows.add(r.getString("tan"));
                                }
                            } catch (SQLException t) {
                                logger.log(Level.SEVERE, t.getMessage(), t);
                            }
                        });
            }
        } catch (Exception e) {
            logger.log(Level.SEVERE, e.getMessage(), e);
        }
        return rows;
    }

    private Boolean invalidateTAN(String owner, String tan) {
        try {
            if (isTANValid(owner, null, null, tan) != null) {
                Timestamp deleted = new Timestamp(Instant.now().toEpochMilli());
                insertStatement("update tan set \"deleted\" = :deleted where \"tan\" = :tan and \"owner\" = :owner",
                        Map.of("deleted", deleted, "tan", tan, "owner", owner));
                return true;
            } else {
                logger.log(Level.WARNING, "TAN already invalidated: owner=" + owner + " tan=" + tan);
                return false;
            }
        } catch (Exception e) {
            logger.log(Level.SEVERE, e.getMessage(), e);
            return false;
        }
    }

    public class SLNPCreateTAN extends Command<CommandServer> {

        public SLNPCreateTAN() {
        }

        @Override
        public List<String> execute(Map<String, Object> params) {
            String isil = (String) params.get("ISIL");
            int anzahl = Integer.parseInt((String) params.get("Anzahl"));
            String wert = (String) params.get("Wert");
            String gruppe = (String) params.get("Gruppe");
            try {
                List<String> tans = createTAN(isil, wert, gruppe, anzahl);
                if (!tans.isEmpty()) {
                    List<String> res = new ArrayList<>();
                    res.add("600 OK. CreateTAN Done");
                    for (String tan : tans) {
                        res.add("601 TAN:" + tan);
                    }
                    res.add("250 SLNPEndOfData");
                    res.add("");
                    return res;
                } else {
                    String errorMessage = "database did not return any TAN";
                    logger.log(Level.SEVERE, errorMessage);
                    return Arrays.asList("510 NotOkCode:" + errorMessage, "");
                }
            } catch (Throwable t) {
                String errorMessage = t.getMessage();
                logger.log(Level.SEVERE, errorMessage, t);
                return Arrays.asList("510 NotOkCode:" + errorMessage, "");
            }
        }
    }

    public class SLNPInsertTAN extends Command<CommandServer> {

        public SLNPInsertTAN() {
        }

        @Override
        public List<String> execute(Map<String, Object> params) {
            String isil = (String) params.get("ISIL");
            String wert = (String) params.get("Wert");
            String gruppe = (String) params.get("Gruppe");
            String tan = (String) params.get("TAN");
            try {
                if (!RandomTAN.validateTAN(tan)) {
                    return Arrays.asList("501 TanServ: TAN has no valid checksum", "");
                }
                if (insertTAN(isil, wert, gruppe, tan)) {
                    return Arrays.asList("600 OK. InsertTAN Done",
                            "601 TAN:" + tan, "250 SLNPEndOfData", "");
                } else {
                    return Arrays.asList("510 NotOkCode:unable to insert, see server log:" + tan, "");
                }
            } catch (Throwable t) {
                String errorMessage = t.getMessage();
                logger.log(Level.SEVERE, errorMessage, t);
                return Arrays.asList("510 NotOkCode:" + errorMessage, "");
            }
        }
    }

    public class SLNPShowTAN extends Command<CommandServer> {

        public SLNPShowTAN() {
        }

        @Override
        public List<String> execute(Map<String, Object> params) {
            String isil = (String) params.get("ISIL");
            String tan = (String) params.get("TAN");
            try {
                Map<String, Object> row = showTAN(isil, tan);
                if (row != null && !row.isEmpty()) {
                    return Arrays.asList(
                            "600 OK",
                            "601 TAN:" + row.get("tan"),
                            "601 ISIL:" + row.get("owner"),
                            "601 Wert:" + row.get("value"),
                            "601 Gruppe:" + row.get("grp"),
                            "601 Create-Datum:" + row.get("created"),
                            "601 Delete-Datum:" + row.get("deleted"),
                            "250 SLNPEndOfData",
                            ""
                    );
                } else {
                    return Arrays.asList("501 TanServ: TAN Not Found", "");
                }
            } catch (Throwable t) {
                String errorMessage = t.getMessage();
                logger.log(Level.SEVERE, errorMessage, t);
                return Arrays.asList("510 NotOkCode:" + errorMessage, "");
            }
        }
    }

    public class SLNPShowTanGroup extends Command<CommandServer> {

        public SLNPShowTanGroup() {
        }

        @Override
        public List<String> execute(Map<String, Object> params) {
            String isil = (String) params.get("ISIL");
            String wert = (String) params.get("Wert");
            String gruppe = (String) params.get("Gruppe");
            try {
                List<String> tans = getAllValidTANs(isil, wert, gruppe);
                if (!tans.isEmpty()) {
                    List<String> res = new ArrayList<>();
                    res.add("600 OK");
                    for (String tan : tans) {
                        res.add("601 TAN:" + tan);
                    }
                    res.add("250 SLNPEndOfData");
                    res.add("");
                    return res;
                } else {
                    return Arrays.asList("501 TanServ: TAN Not Found", "");
                }
            } catch (Throwable t) {
                String errorMessage = t.getMessage();
                logger.log(Level.SEVERE, errorMessage, t);
                return Arrays.asList("510 NotOkCode:" + errorMessage, "");
            }
        }
    }

    public class SLNPIsTanValid extends Command<CommandServer> {

        public SLNPIsTanValid() {
        }

        @Override
        public List<String> execute(Map<String, Object> params) {
            String isil = (String) params.get("ISIL");
            String wert = (String) params.get("Wert");
            String gruppe = (String) params.get("Gruppe");
            String tan = (String) params.get("TAN");
            try {
                if (tan != null && tan.length() > 6 && !RandomTAN.validateTAN(tan)) {
                    return Arrays.asList("501 TanServ: TAN has no valid checksum", "");
                }
                Map<String, Object> row = isTANValid(isil, wert, gruppe, tan);
                if (row != null && row.containsKey("created") && !row.containsKey("deleted")) {
                    return Arrays.asList(
                            "600 OK. TAN is Valid",
                            "601 Wert:" + row.get("value"),
                            "601 Create-Datum:" + row.get("created"),
                            "601 Gruppe:" + row.get("grp"),
                            "250 SLNPEndOfData",
                            "");
                } else {
                    return Arrays.asList("501 TanServ: TAN Not Valid", "");
                }
            } catch (Throwable t) {
                String errorMessage = t.getMessage();
                logger.log(Level.SEVERE, errorMessage, t);
                return Arrays.asList("510 NotOkCode:" + errorMessage, "");
            }
        }
    }

    public class SLNPDeleteTAN extends Command<CommandServer> {

        public SLNPDeleteTAN() {
        }

        @Override
        public List<String> execute(Map<String, Object> params) {
            String isil = (String) params.get("ISIL");
            String tan = (String) params.get("TAN");
            try {
                if (invalidateTAN(isil, tan)) {
                    return Arrays.asList("240 OK. DeleteTAN Done", "");
                } else {
                    return Arrays.asList("501 TanServ: TAN could not be deleted", "");
                }
            } catch (Throwable t) {
                String errorMessage = t.getMessage();
                logger.log(Level.SEVERE, errorMessage, t);
                return Arrays.asList("510 NotOkCode:" + errorMessage, "");
            }
        }
    }
}
